from pathlib import Path
from typing import Iterator
import polars as pl

def _parse_single_root(root_lines: list[str]) -> dict:
    """Parse a single root block into a dictionary.
    """
    # Parse header line
    header = root_lines[0].strip()
    state_num = int(header[5:header.index(':')])
    parts = header.strip().split()
    energy_ev = float(parts[5])
    energy_cm = float(parts[7])
    spin_expectation_value = float(parts[11])
    multiplicity = int(parts[13])

    # Parse root contributions
    orbital_transitions = []
    for line in root_lines[1:]:
        parts = line.replace("->", "").replace(":", "").replace(")", "").strip().split()
        orbital_transitions.append({
            'Donor': str(parts[0]),
            'Acceptor': str(parts[1]),
            'Weight': float(parts[2]),
        })

    return {
        'Root': state_num,
        'Energy [eV]': energy_ev,
        'Energy [1/cm]': energy_cm,
        '<S**2>': spin_expectation_value,
        'Multiplicity': multiplicity,
        'Orbital Transitions': orbital_transitions
    }

def _iter_roots(lines: list[str]) -> Iterator[list[str]]:
    """Iterate over root blocks in the input text."""
    current_state = []

    for i, line in enumerate(lines):
        line = line.strip()
        if not line:
            if lines[i+1].strip().startswith("STATE"):
                continue
            else:
                break

        if line.startswith('STATE'):
            if current_state:
                yield current_state
            current_state = [line]
        elif current_state:
            current_state.append(line)

    if current_state:  # Don't forget the last state
        yield current_state


def roots(orca_output: Path) -> pl.DataFrame:
    """Parse roots from ORCA output into a Polars DataFrame."""
    # Read file content
    content = Path(orca_output).read_text().splitlines()

    # Find the start of root section(s)
    start_indicies = []
    for line in content:
        if "TD-DFT/TDA EXCITED STATES" in line or "TD-DFT EXCITED STATES" in line:
            start_indicies.append(content.index(line))
            # Find the custom offset (different for different tables..)
            while not content[start_indicies[-1]].startswith("STATE"):
                start_indicies[-1] += 1

    # Parse roots
    roots = []
    for start_idx in start_indicies:
        for root_lines in _iter_roots(content[start_idx:]):
            root_data = _parse_single_root(root_lines)
            roots.append(root_data)

    # Flatten the data structure
    flattened_roots = []
    for root in roots:
        for orb_contrib in root['Orbital Transitions']:
            flattened_roots.append({
                'Root': root['Root'],
                'Multiplicity': root['Multiplicity'],
                'Donor': orb_contrib['Donor'],
                'Acceptor': orb_contrib['Acceptor'],
                'Weight': orb_contrib['Weight'],
                'Energy [1/cm]': root['Energy [1/cm]'],
            })

    # Create and transform DataFrame
    df = pl.DataFrame(flattened_roots)
    return df.sort('Root', 'Weight', descending=[False, True])



from pathlib import Path
from typing import Iterator
import polars as pl

def _parse_single_state(state_lines: list[str]) -> dict:
    """Parse a single SOC state block into a dictionary.

    Example input format:
    STATE 1: 0.000000
       1.000000    1.000000    0.000000     1    1    1
       0.000000    0.000000    0.000000     2    1    0
    """
    # Parse header line
    header = state_lines[0].strip()
    state_num = int(header[5:header.index(':')])
    energy = float(header[header.index(':') + 1:])

    # Parse root contributions
    roots = []
    for line in state_lines[1:]:
        parts = line.replace(":", "").strip().split()
        roots.append({
            'Weight': float(parts[0]),
            'Real': float(parts[1]),
            'Imag': float(parts[2]),
            'Root': int(parts[3]),
            'Spin': int(parts[4]),
            'Ms': int(parts[5])
        })

    return {
        'State': state_num,
        'Energy [1/cm]': energy,
        'roots': roots
    }

def _iter_soc_states(lines: list[str]) -> Iterator[list[str]]:
    """Iterate over SOC state blocks in the input text."""
    current_state = []

    for line in lines:
        line = line.strip()
        if not line:
            break

        if line.startswith('STATE'):
            if current_state:
                yield current_state
            current_state = [line]
        elif current_state:
            current_state.append(line)

    if current_state:  # Don't forget the last state
        yield current_state

def soc_states(orca_output: Path) -> pl.DataFrame:
    """Parse SOC states from ORCA output into a Polars DataFrame."""
    # Read file content
    content = Path(orca_output).read_text().splitlines()

    # Find the start of SOC states section
    try:
        start_idx = next(i for i, line in enumerate(content)
                        if "Eigenvectors of the SOC matrix:" in line.strip()) + 3
    except StopIteration:
        raise ValueError("SOC matrix section not found in input")

    # Parse states
    soc_states = []
    for state_lines in _iter_soc_states(content[start_idx:]):
        state_data = _parse_single_state(state_lines)

        # Flatten the data structure
        for root in state_data['roots']:
            soc_states.append({
                'State': state_data['State'],
                'Energy [1/cm]': state_data['Energy [1/cm]'],
                **root
            })

    # Create and transform DataFrame
    df = pl.DataFrame(soc_states)
    return (df.group_by(['State', 'Spin'])
              .agg([
                  pl.first('Root'),
                  pl.sum('Weight'),
                  pl.first('Energy [1/cm]')
              ])
              .sort(['State', 'Weight'], descending=[False, True]))



from pathlib import Path

import polars as pl

def ir_spectrum(orca_output: Path) -> pl.DataFrame:
    lines = Path(orca_output).read_text().splitlines()

    TABLE_HEADER = "IR SPECTRUM"
    TABLE_HEADER_OFFSET = 5

    table_start_idx = next(i for i, line in enumerate(lines) if TABLE_HEADER in line.strip())
    table_start_idx += TABLE_HEADER_OFFSET

    # Collect table
    rows = []
    for row in lines[table_start_idx:]:
        # Stop on empty line
        if not row.strip():
            break

        rows.append(row)

    processed_rows = [row.replace(":", "").replace("(", "").replace(")", "").strip().split() for row in rows]

    df = pl.DataFrame(
        processed_rows,
        orient='row',
        schema={
            "Mode": pl.Int64,
            "Frequency [1/cm]": pl.Float64,
            "Epsilon [L/(mol*cm)]": pl.Float64,
            "Intensity [km/mol]": pl.Float64,
            "T2": pl.Float64,
            "TX": pl.Float64,
            "TY": pl.Float64,
            "TZ": pl.Float64
        }
    )

    df = df.with_columns(
        (pl.col("Intensity [km/mol]") / pl.col("Intensity [km/mol]").max()).alias("Rel. Intensity")
    )

    return df



from pathlib import Path

import polars as pl

def soc_absorption_spectrum(orca_output: Path) -> pl.DataFrame:
    lines = Path(orca_output).read_text().splitlines()

    TABLE_HEADER = "SOC CORRECTED ABSORPTION SPECTRUM VIA TRANSITION ELECTRIC DIPOLE MOMENTS"
    TABLE_HEADER_OFFSET = 5

    table_start_idx = next(i for i, line in enumerate(lines) if TABLE_HEADER in line.strip())
    table_start_idx += TABLE_HEADER_OFFSET

    # Collect table
    rows = []
    for row in lines[table_start_idx:]:
        # Stop on empty line
        if not row.strip():
            break

        rows.append(row)

    # Process table
    processed_rows = []
    for row in rows:
        row = row.replace("A", "").replace("B", "").replace("->", "")
        to_state, to_spin = row.split()[1].split("-")
        processed_row = [to_state, to_spin] + row.split()[2:]
        processed_rows.append(processed_row)

    df = pl.DataFrame(
        processed_rows,
        orient="row",
        schema={
            "State": pl.Int64,
            "Multiplicity": pl.Float64,
            "Energy [ev].": pl.Float64,
            "Energy [1/cm].": pl.Float64,
            "Wavelength [nm]": pl.Float64,
            "Osc. Strength": pl.Float64,
            "D2": pl.Float64,
            "DX": pl.Float64,
            "DY": pl.Float64,
            "DZ": pl.Float64
        }
    )

    df = df.with_columns(
            (pl.col("Osc. Strength") / pl.col("Osc. Strength").max()).alias("Rel. Intensity")
        )

    return df
